
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.contentsprocessing;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipInputStream;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.zipentryprocessing.ZipEntryProcessor;
import de.unkrig.commons.io.UnclosableInputStream;
import de.unkrig.commons.lang.protocol.Predicate;

/**
 * @see #process(String, InputStream)
 */
public
class ZipContentsProcessor implements ContentsProcessor {

    private final Predicate<String>             nestedZipPathPredicate;
    private final ZipEntryProcessor             nestedZipEntryProcessor;
    private final ContentsProcessor             defaultContentsProcessor;
    private final ExceptionHandler<IOException> exceptionHandler;

    /**
     * @see #process(String, InputStream)
     */
    public
    ZipContentsProcessor(
        Predicate<String> nestedZipPathPredicate,
        ZipEntryProcessor nestedZipEntryProcessor,
        ContentsProcessor defaultContentsProcessor
    ) {
        this.nestedZipPathPredicate = nestedZipPathPredicate;
        this.nestedZipEntryProcessor = nestedZipEntryProcessor;
        this.defaultContentsProcessor = defaultContentsProcessor;
        this.exceptionHandler = ExceptionHandler.defaultHandler();
    }

    /**
     * @see #process(String, InputStream)
     */
    public
    ZipContentsProcessor(
        Predicate<String>             nestedZipPathPredicate,
        ZipEntryProcessor             nestedZipEntryProcessor,
        ContentsProcessor             defaultContentsProcessor,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        this.nestedZipPathPredicate = nestedZipPathPredicate;
        this.nestedZipEntryProcessor = nestedZipEntryProcessor;
        this.defaultContentsProcessor = defaultContentsProcessor;
        this.exceptionHandler = exceptionHandler;
    }

    /**
     * If the contents of {@code is} is not in ZIP format, then it is transformed through {code
     * defaultContentsProcessor}. Otherwise, if {@code nestedZipPathPredicate} evaluates to {@code false}, nothing
     * is done. Otherwise is opened as a {@link ZipInputStream}, and for each ZIP entry the {@code
     * nestedZipEntryProcessor} is invoked.
     */
    @Override public void
    process(String name, InputStream is) throws IOException {

        // Make the stream resettable.
        if (!is.markSupported()) is = new BufferedInputStream(is);
        is.mark(4);

        // Check if the stream contents starts with the magic number.
        boolean isZip;
        {
            byte[] buf = new byte[4];
            try {
                isZip = (is.read(buf) == 4 && buf[3] == 0x04 && buf[2] == 0x03 && buf[1] == 0x4B && buf[0] == 0x50);
            } catch (IOException ioe) {
                throw wrap("Reading the first few bytes from '" + name + "'", ioe);
            } catch (RuntimeException re) {
                throw wrap("Reading the first few bytes from '" + name + "'", re);
            }
            try { is.reset(); } catch (Exception e) {}
        }

        if (!isZip) {
            this.defaultContentsProcessor.process(name, is);
            return;
        }

        if (!this.nestedZipPathPredicate.evaluate(name + '!')) return;

        // We need to close the ZIS later (to release the resources allocated by the ZIP inflater), but the contract
        // of 'process()' requires that the underlying 'is' must not be closed.
        ZipInputStream zis = new ZipInputStream(new UnclosableInputStream(is));
        try {

            ZipEntry entry;
            try {
                entry = zis.getNextEntry();
            } catch (ZipException e) {

                // In rare cases, we get a 'ZipException: invalid entry CRC (expected 0x0 but got 0xfde2c118)' -
                // which obviously means that we hit a binary file with the correct magic number, but an
                // otherwise invalid first entry.
                is.reset();
                this.defaultContentsProcessor.process(name, is);
                return;
            } catch (IOException ioe) {
                throw wrap("Reading first entry of nested ZIP '" + name + "'", ioe);
            } catch (RuntimeException re) {
                throw wrap("Reading first entry of nested ZIP '" + name + "'", re);
            }

            if (entry == null) {

                // This should not happen, because it indicates a wrong magic number, and the magic number
                // has been checked before.
                is.reset();
                this.defaultContentsProcessor.process(name, is);
                return;
            }

            do {
                String entryName = name + '!' + entry.getName();

                this.nestedZipEntryProcessor.process(entry, entryName, zis);

                try {
                    entry = zis.getNextEntry();
                } catch (IOException ioe) {
                    this.exceptionHandler.handle(wrap("Reading next entry after '" + entryName + "'", ioe));
                    return;
                } catch (RuntimeException re) {
                    this.exceptionHandler.handle(wrap("Reading next entry after '" + entryName + "'", re));
                    return;
                }
            } while (entry != null);

            zis.close();
        } finally {
            try { zis.close(); } catch (Exception e) {}
        }
    }
}
