
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import de.unkrig.commons.file.FileUtil;

/**
 * @see #transform(File, File)
 */
public
class InPlaceFileTransformer implements FileTransformer {

    private final FileTransformer delegate;
    private final boolean         keepOriginals;

    /**
     * If {@code keepOriginals} is {@code true}, then for each in-place file transformation (see {@link
     * #transform(File, File)} a copy of the original file with a similar name is retained.
     */
    public
    InPlaceFileTransformer(FileTransformer delegate, boolean keepOriginals) {
        this.delegate      = delegate;
        this.keepOriginals = keepOriginals;
    }

    /**
     * If {@code in} and {@code out} denote different files (or directories), then all missing parent directories for
     * {@code out} are created, and the {@code delegate} is invoked.
     * <p>
     * Otherwise, the {@code in} file (or directory) is first transformed into a temporary file, which is then moved
     * over the original file (as atomically as possible). This is called an "in-place file transformation".
     * Iff the {@code keepOriginals} flag is set, then the original file is renamed before the temporary file is
     * renamed to the original name.
     */
    @Override public void
    transform(File in, File out) throws IOException {
        if (in.equals(out)) {
            this.transformInPlace(in);
        } else {
            transformOutOfPlace(in, out);
        }
    }

    private void
    transformInPlace(File file) throws IOException {

        File newFile = new File(file.getParentFile(), '.' + file.getName() + ".new");
        if (newFile.exists() && !newFile.delete()) {
            throw new IOException("File '" + newFile + "' is in the way and cannot be deleted");
        }

        try {
            this.delegate.transform(file, newFile);
        } catch (IOException ioe) {
            FileUtil.attemptToDeleteRecursively(newFile);
            throw wrap("Transforming in-place to '" + newFile + "'", ioe);
        } catch (RuntimeException re) {
            FileUtil.attemptToDeleteRecursively(newFile);
            throw wrap("Transforming in-place to '" + newFile + "'", re);
        }

        if (!newFile.setLastModified(file.lastModified())) {
            throw new IOException("Could not set modification time of '" + newFile + "'");
        }

        if (this.keepOriginals) {
            File origFile = new File(file.getParentFile(), '.' + file.getName() + ".orig");
            if (!origFile.delete()) {
                FileUtil.attemptToDeleteRecursively(newFile);
                throw new IOException("File '" + newFile + "' is in the way and cannot be deleted");
            }
            if (!file.renameTo(origFile)) {
                FileUtil.attemptToDeleteRecursively(newFile);
                throw new IOException("Could not rename '" + file + "' to '" + origFile);
            }
        } else {
            if (!file.delete()) throw new IOException("Could not delete '" + file + "'");
        }

        if (!newFile.renameTo(file)) throw new IOException("Could not rename '" + newFile + "' to '" + file);
    }

    /**
     * If the creation of the {@code out} file fails for any reason, then the {@code out} file is deleted.
     * @param out Must be different from {@code in}
     */
    public void
    transformOutOfPlace(File in, File out) throws FileNotFoundException, IOException {

        // Create parent directories as necessary.
        {
            File outDir = out.getParentFile();
            if (outDir != null && !outDir.exists() && !outDir.mkdirs()) {
                throw new IOException("Cannot create output directory '" + outDir + "'");
            }
        }

        this.delegate.transform(in, out);

        if (!out.setLastModified(in.lastModified())) {
            throw new IOException("Could not set modification time of output file");
        }
    }
}
