
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2012, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.lang.protocol;

import java.util.Collection;

public
class RunnableUtil {

    private RunnableUtil() {}

    /**
     * Runs {@code delegate1}, then {@code delegate2} in the current thread.
     *
     * @return    A runnable that runs {@code delegate1} and then {@code delegate2}
     * @throws EX The exception thrown by {@code delegate1} or {@code delegate2}; if thrown by {@code delegate1}, then
     *            {@code delegate2} is not run
     *
     * @see RunnableWhichThrows#run()
     */
    public static <EX extends Exception> RunnableWhichThrows<EX>
    runSequentially(final RunnableWhichThrows<EX> delegate1, final RunnableWhichThrows<EX> delegate2) {
        return new RunnableWhichThrows<EX>() {

            @Override public void
            run() throws EX {
                delegate1.run();
                delegate2.run();
            }
        };
    }

    /**
     * Runs the given {@code delegates} in the current thread, in the order of the collection's iterator.
     *
     * @return    A runnable that runs all the {@code delegates} in strict sequence
     * @throws EX The exception thrown by one of the {@code delegates}; the following runnables are not run
     *
     * @see RunnableWhichThrows#run()
     */
    public static <EX extends Exception> RunnableWhichThrows<EX>
    runSequentially(final Collection<RunnableWhichThrows<EX>> delegates) {
        return new RunnableWhichThrows<EX>() {

            @Override public void
            run() throws EX {
                for (RunnableWhichThrows<EX> delegate : delegates) {
                    delegate.run();
                }
            }
        };
    }

    /**
     * Converts a {@link RunnableWhichThrows} into a different {@link RunnableWhichThrows}, which is possible iff
     * the source runnable's exception is a subclass of the target runnable's exception.
     *
     * @param <EX1> The source runnable's exception
     * @param <EX2> The target runnable's exception
     */
    public static <EX2 extends Exception, EX1 extends EX2> RunnableWhichThrows<EX2>
    cast(final RunnableWhichThrows<EX1> delegate) {
        return new RunnableWhichThrows<EX2>() {
            @Override public void run() throws EX2 { delegate.run(); }
        };
    }

    /**
     * Converts a {@link Runnable} into a {@link RunnableWhichThrows}.
     */
    public static <EX extends Exception> RunnableWhichThrows<EX>
    cast(final Runnable delegate) {
        return new RunnableWhichThrows<EX>() {
            @Override public void run() { delegate.run(); }
        };
    }

    /**
     * Returns a runnable that invokes the delegate iff the current time is at or after the given expiration time.
     */
    public static Runnable
    notBefore(final long expirationTime, final Runnable delegate) {
        return new Runnable() {
            @Override public void run() { if (System.currentTimeMillis() >= expirationTime) delegate.run(); }
        };
    }

    /**
     * Returns a runnable which invokes the delegate once when it is first run, and after that iff the delegate was
     * last run at least a given interval ago.
     */
    public static Runnable
    atMostEvery(final long interval, final Runnable delegate) {
        return new Runnable() {

            private long expirationTime;

            @Override public synchronized void
            run() {
                if (System.currentTimeMillis() >= this.expirationTime) {
                    this.expirationTime = System.currentTimeMillis() + interval;
                    delegate.run();
                }
            }
        };
    }
}
