
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.text.expression;

import static de.unkrig.commons.text.expression.Scanner.TokenType.CHARACTER_LITERAL;
import static de.unkrig.commons.text.expression.Scanner.TokenType.FLOATING_POINT_LITERAL;
import static de.unkrig.commons.text.expression.Scanner.TokenType.IDENTIFIER;
import static de.unkrig.commons.text.expression.Scanner.TokenType.INTEGER_LITERAL;
import static de.unkrig.commons.text.expression.Scanner.TokenType.STRING_LITERAL;

import java.io.Reader;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.regex.Pattern;

import de.unkrig.commons.lang.ExceptionUtil;
import de.unkrig.commons.lang.StringUtil;
import de.unkrig.commons.lang.protocol.ProducerWhichThrows;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.expression.Scanner.TokenType;
import de.unkrig.commons.text.parser.AbstractParser;
import de.unkrig.commons.text.parser.ParseException;
import de.unkrig.commons.text.pattern.Glob;
import de.unkrig.commons.text.scanner.AbstractScanner.Token;
import de.unkrig.commons.text.scanner.ScanException;
import de.unkrig.commons.text.scanner.ScannerUtil;

/**
 * Parses an expression like
 * <pre>
 * s == "abc" && (c == 'b' || !b == true)
 * </pre>
 *
 * Supported operators (in ascending priority) are:
 * <dl>
 *   <dt>{@code a ? b : c}</dt>
 *   <dd>
 *     {@code b} if {@code a}, {@link ExpressionEvaluator#toBoolean(Object) converted to boolean}, is true, otherwise
 *     {@code c}.
 *   </dd>
 *
 *   <dt>{@code a || b}</dt>
 *   <dd>
 *     {@code a} if {@code a}, {@link ExpressionEvaluator#toBoolean(Object) converted to boolean}, is true, otherwise
 *     {@code b}. (Equivalent to '{@code a ? a : b}', except that {@code a} is evaluated only once.)
 *   </dd>
 *
 *   <dt>{@code a && b}</dt>
 *   <dd>
 *     {@code b} if {@code a}, {@link ExpressionEvaluator#toBoolean(Object) converted to boolean}, is true, otherwise
 *     {@code false}. (Equivalent to '{@code a ? b : false}'.)
 *   </dd>
 *
 *   <dt>{@code == != < <= > >=}</dt>
 *   <dd>
 *     Arithmetic comparison (if both operands are numeric) or lexicographic comparison (otherwise).
 *   </dd>
 *
 *   <dt>{@code =*}</dt>
 *   <dd>
 *     Wildcard matching/replacing (see {@link Glob}), evaluates to the replacement.
 *   </dd>
 *
 *   <dt>{@code =~}</dt>
 *   <dd>
 *     Regex matching/replacing (see {@link Pattern}), evaluates to the replacement.
 *   </dd>
 *
 *   <dt>{@code + -}</dt>
 *   <dd>
 *     Addition (if both operands are numeric) or string concatenation, subtraction.
 *   </dd>
 *
 *   <dt>{@code * / %}</dt>
 *   <dd>
 *     Multiplication, division, remainder (modulo).
 *   </dd>
 *
 *   <dt>{@code a instanceof reference-type}</dt>
 *   <dd>{@code true} iff {@code a} is non-null and a subtype of {@code reference-type}</dd>
 *
 *   <dt>{@code (...) ! -}</dt>
 *   <dd>
 *     Grouping, logical negation, arithmetic negation.
 *   </dd>
 * </dl>
 * Primaries are:
 * <dl>
 *   <dt>{@code "abc"}</dt>
 *   <dd>String literal</dd>
 *
 *   <dt>{@code 123}</dt>
 *   <dd>Integer literal</dd>
 *
 *   <dt>{@code 123L 123l}</dt>
 *   <dd>Long literal</dd>
 *
 *   <dt>{@code 1F 1.2E+3f}</dt>
 *   <dd>Float literal</dd>
 *
 *   <dt>{@code 1D 1d 1.1 1.1e-3}</dt>
 *   <dd>Double literal</dd>
 *
 *   <dt>{@code true false null}</dt>
 *   <dd>Special constant values</dd>
 *
 *   <dt>{@code abc}</dt>
 *   <dd>Variable reference, see {@link Expression#evaluate(Object[])}</dd>
 *
 *   <dt>{@code new pkg.Clazz(arg1, arg2)}<br>{@code new RootPackageClazz(arg1, arg2)}<br>{@code new ImportedClazz(arg1, arg2)}</dt>
 *   <dd>Class instance creation</dd>
 *
 *   <dt>{@code new pkg.Clazz}<br>{@code pkg.Clazz}<br>{@code pkg.Clazz()}</dt>
 *   <dd>Abbreviated forms of {@code new pkg.Clazz()}</dd>
 *
 *   <dt>{@code new pkg.Clazz[x]}<br>{@code new pkg.Clazz[x][y][][]}<br>{@code new int[x]}<br>{@code new int[x][y][][]}</dt>
 *   <dd>Array creation</dd>
 *
 *   <dt>{@code pkg.Clazz[x]}<br>{@code pkg.Clazz[x][y][][]}<br>{@code int[x]}<br>{@code int[x][y][][]}</dt>
 *   <dd>Abbreviated array creation</dd>
 *
 *   <dt>{@code x.Name}</dt>
 *   <dd>Attribute reference, invokes 'getName()'</dd>
 *
 *   <dt>{@code x.meth(a, b, c)}</dt>
 *   <dd>Method invocation</dd>
 * </dl>
 * {@code Null} operands are handled leniently, e.g. "7 == null" is false, "null == null" is true.
 *
 * <p>To implement your parser, derive from this class and implement the abstract methods.
 *
 * @param T The type returned by {@link #parse()}
 * @param E The exception thrown by the abstract handlers (e.g. {@link #conditional(Object, Object, Object)})
 */
public abstract
class Parser<T, E extends Exception> extends AbstractParser<TokenType> {

    public
    Parser(ProducerWhichThrows<Token<TokenType>, ScanException> tokenProducer) {
        super(tokenProducer);
    }

    public
    Parser(Reader in) {
        this(ScannerUtil.toDocumentScanner(Scanner.stringScanner(), in));
    }

    public
    Parser(String expression) {
        this(Scanner.stringScanner().setInput(expression));
    }

    public String[]
    getImports() {
        return this.imports.clone();
    }

    /**
     * By default, there is only one import: "java.lang".
     */
    public Parser<T, E>
    setImports(String[] imports) {
        this.imports = imports.clone();
        return this;
    }

    public ClassLoader
    getClassLoader() {
        return this.classLoader;
    }

    /**
     * @param classLoader Used to load classes named in the expression; by default the class loader which loaded
     *                    the {@link Parser} class.
     */
    public Parser<T, E>
    setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
        return this;
    }

    /**
     * Enable one of the parser extensions listed in {@link Parser.Extension}. By default, all extensions are enabled.
     */
    public void
    enableExtension(Extension extension) { this.extensions.add(extension); }

    /**
     * Disable one of the parser extensions listed in {@link Parser.Extension}. By default, all extensions are enabled.
     */
    public void
    disableExtension(Extension extension) { this.extensions.remove(extension); }

    public T
    parse() throws ParseException, E {
        try {
            T result = this.parseExpression().toValue();
            this.eoi();
            return result;
        } catch (ParseException pe) {
            throw ExceptionUtil.wrap("At " + this.scanner.toString(), pe);
        } catch (RuntimeException re) {
            throw ExceptionUtil.wrap("At " + this.scanner.toString(), re);
        } catch (Exception e) {
            @SuppressWarnings("unchecked") E ee = (E) e;
            throw ExceptionUtil.wrap("At " + this.scanner.toString(), ee);
        }
    }

    public enum UnaryOperator {
        MINUS             ("-"),
        LOGICAL_COMPLEMENT("!"),
        BITWISE_COMPLEMENT("~");

        private final String text;
        UnaryOperator(String text) { this.text = text; }
        @Override public String toString() { return this.text; }
    }

    public enum BinaryOperator {
        EQUAL        ("=="),
        GLOB         ("=*"),
        REGEX        ("=~"),
        NOT_EQUAL    ("!="),
        LESS         ("<"),
        LESS_EQUAL   ("<="),
        GREATER      (">"),
        GREATER_EQUAL(">="),
        PLUS         ("+"),
        MINUS        ("-"),
        MULTIPLY     ("*"),
        DIVIDE       ("/"),
        MODULO       ("%"),
        LOGICAL_OR   ("||"),
        LOGICAL_AND  ("&&"),
        BITWISE_OR   ("|"),
        BITWISE_XOR  ("^"),
        BITWISE_AND  ("&"),
        LEFT_SHIFT   ("<<"),
        RIGHT_SHIFT  (">>"),
        RIGHT_USHIFT (">>>");

        private final String text;
        BinaryOperator(String text) { this.text = text; }
        @Override public String toString() { return this.text; }
    }

    /**
     * Various extensions to the Java expression syntax.
     *
     * @see Parser#enableExtension(Extension)
     * @see Parser#disableExtension(Extension)
     */
    enum Extension {

        /** The {@code new} keyword can be left out when a class object is instantiated. */
        NEW_CLASS_WITHOUT_KEYWORD,

        /** The "()" can be omitted when instantiating a class object with its zero-arg constructor. */
        NEW_CLASS_WITHOUT_PARENTHESES,

        /** The operator '=*' implements a wildcard pattern matching algorithm, e.g.<br>{@code "foo.java" =* "*.java"} */
        OPERATOR_GLOB,

        /** The operator '=~' implements a regex pattern matching algorithm, e.g.<br>{@code ".*" =~ "abc"} */
        OPERATOR_REGEX,
    }

    private String[]                 imports = new String[] { "java.lang" };
    private ClassLoader              classLoader = this.getClass().getClassLoader();
    private final EnumSet<Extension> extensions = EnumSet.allOf(Extension.class);

    /**
     * It is generally difficult to distinguish between values, types and packages, e.g. "a.b.c" could represent
     * <ul>
     *   <li>Field c of field b of variable a (a value)
     *   <li>Class or interface c in package a.b (a type)
     *   <li>Package a.b.c
     * </ul>
     * This is why many of the productions (corresponding with the various {@code parse...()} methods) return an
     * {@link Atom}.
     */
    interface Atom<T, E extends Exception> {

        /**
         * @return                This atom's value (where {@code null} is a perfectly legal value)
         * @throws ParseException Iff this atom does not pose a value, e.g. "String[]"
         */
        T toValue() throws E, ParseException;

        /**
         * @return {@code null} iff this atom does not pose a type, e.g. "1 + 2"
         */
        @Nullable Class<?> toType();

        /**
         * @return {@code null} iff this atom does not pose a package, e.g. "1 + 2"
         */
        @Nullable String toPackage();
    }

    /**
     * @return An {@link Atom} that poses a value
     */
    private static <T, E extends Exception> Atom<T, E>
    value(final T t) {
        return new Atom<T, E>() {
            @Override public T                  toValue()   { return t; }
            @Override public @Nullable Class<?> toType()    { return null; }
            @Override public @Nullable String   toPackage() { return null; }
        };
    }

    /**
     * @return An {@link Atom} that poses a type
     */
    private Atom<T, E>
    type(final Class<?> type) {
        return new Atom<T, E>() {

            @Override public T
            toValue() throws E, ParseException {
                if (
                    Parser.this.extensions.contains(Extension.NEW_CLASS_WITHOUT_KEYWORD)
                    && Parser.this.extensions.contains(Extension.NEW_CLASS_WITHOUT_PARENTHESES)
                ) {
                    return Parser.this.newClass(type, Collections.<T>emptyList());
                } else {
                    throw new ParseException("'" + type.getName() + "' is a type, not a value");
                }
            }

            @Override public Class<?>         toType()    { return type; }
            @Override public @Nullable String toPackage() { return null; }
        };
    }

    /**
     * <pre>
     * expression := conditional
     * </pre>
     */
    private Atom<T, E>
    parseExpression() throws ParseException, E {
        return this.parseConditional();
    }

    /**
     * <pre>
     * conditional := logical-or [ '?' logical-or ':' conditional ]
     * </pre>
     */
    private Atom<T, E>
    parseConditional() throws ParseException, E {
        Atom<T, E> lhs = this.parseLogicalOr();
        if (!this.peekRead("?")) return lhs;

        T mhs = this.parseLogicalOr().toValue();
        this.read(":");
        T rhs = this.parseConditional().toValue();
        return value(this.conditional(lhs.toValue(), mhs, rhs));
    }

    /**
     * <pre>
     * logical-or := logical-and { '||' logical-and }
     * </pre>
     */
    private Atom<T, E>
    parseLogicalOr() throws ParseException, E {
        Atom<T, E> lhs = this.parseLogicalAnd();
        while (this.peekRead("||")) {
            lhs = value(this.binaryOperation(
                lhs.toValue(),
                BinaryOperator.LOGICAL_OR,
                this.parseLogicalAnd().toValue()
            ));
        }
        return lhs;
    }

    /**
     * <pre>
     * logical-and := bitwise-or { '&&' bitwise-or }
     * </pre>
     */
    private Atom<T, E>
    parseLogicalAnd() throws ParseException, E {
        Atom<T, E> lhs = this.parseBitwiseOr();
        while (this.peekRead("&&")) {
            lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.LOGICAL_AND, this.parseBitwiseOr().toValue()));
        }
        return lhs;
    }

    /**
     * <pre>
     * bitwise-or := bitwise-xor { '|' bitwise-xor }
     * </pre>
     */
    private Atom<T, E>
    parseBitwiseOr() throws ParseException, E {
        Atom<T, E> lhs = this.parseBitwiseXor();
        while (this.peekRead("|")) {
            lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.BITWISE_OR, this.parseBitwiseXor().toValue()));
        }
        return lhs;
    }

    /**
     * <pre>
     * bitwise-xor := bitwise-and { '^' bitwise-and }
     * </pre>
     */
    private Atom<T, E>
    parseBitwiseXor() throws ParseException, E {
        Atom<T, E> lhs = this.parseBitwiseAnd();
        while (this.peekRead("^")) {
            lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.BITWISE_XOR, this.parseBitwiseAnd().toValue()));
        }
        return lhs;
    }

    /**
     * <pre>
     * bitwise-and := relational { '^' relational }
     * </pre>
     */
    private Atom<T, E>
    parseBitwiseAnd() throws ParseException, E {
        Atom<T, E> lhs = this.parseRelational();
        while (this.peekRead("&")) {
            lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.BITWISE_AND, this.parseRelational().toValue()));
        }
        return lhs;
    }

    /**
     * <pre>
     * relational :=
     *    shift {
     *       '==' shift
     *       | '=*' shift
     *       | '=~' shift
     *       | '!=' shift
     *       | '<'  shift
     *       | '<=' shift
     *       | '>'  shift
     *       | '>=' shift
     *       | 'instanceof' reference-type
     *    }
     * </pre>
     */
    private Atom<T, E>
    parseRelational() throws ParseException, E {
        Atom<T, E> lhs = this.parseShift();

        for (;;) {
            BinaryOperator operator;
            if (this.extensions.contains(Extension.OPERATOR_GLOB) && this.peekRead("=*")) {
                lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.GLOB, this.parseRelational().toValue()));
            } else
            if (this.extensions.contains(Extension.OPERATOR_REGEX) && this.peekRead("=~")) {
                lhs = value(binaryOperation(lhs.toValue(), BinaryOperator.REGEX, this.parseRelational().toValue()));
            } else
            if ((operator = this.peekReadEnum(
                BinaryOperator.EQUAL, BinaryOperator.NOT_EQUAL, BinaryOperator.LESS, BinaryOperator.LESS_EQUAL,
                BinaryOperator.GREATER, BinaryOperator.GREATER_EQUAL
            )) != null) {
                lhs = value(binaryOperation(lhs.toValue(), operator, this.parseRelational().toValue()));
            } else
            if (this.peekRead("instanceof")) {
                lhs = value(instanceoF(lhs.toValue(), parseReferenceType()));
            } else
            {
                break;
            }

        }
        return lhs;
    }

    /**
     * <pre>
     * shift :=
     *     additive { ( '<<' | '>>' | '>>>' ) additive }
     * </pre>
     */
    private Atom<T, E>
    parseShift() throws ParseException, E {
        Atom<T, E> lhs = parseAdditive();
        for (;;) {
            final BinaryOperator op = this.peekReadEnum(BinaryOperator.LEFT_SHIFT, BinaryOperator.RIGHT_SHIFT, BinaryOperator.RIGHT_USHIFT);
            if (op == null) return lhs;
            lhs = value(binaryOperation(lhs.toValue(), op, parseAdditive().toValue()));
        }
    }

    /**
     * <pre>
     * additive :=
     *     multiplicative { ( '+' | '-' ) multiplicative }
     * </pre>
     */
    private Atom<T, E>
    parseAdditive() throws ParseException, E {
        Atom<T, E> mul = parseMultiplicative();
        for (;;) {
            final BinaryOperator op = this.peekReadEnum(BinaryOperator.PLUS, BinaryOperator.MINUS);
            if (op == null) return mul;
            mul = value(binaryOperation(mul.toValue(), op, parseMultiplicative().toValue()));
        }
    }

    /**
     * <pre>
     * multiplicative :=
     *    instanceof { ( '*' | '/' | '%' ) instanceof }
     * </pre>
     */
    private Atom<T, E>
    parseMultiplicative() throws ParseException, E {
        Atom<T, E> lhs = parseSelector();
        for (;;) {
            final BinaryOperator op = this.peekReadEnum(
                BinaryOperator.MULTIPLY,
                BinaryOperator.DIVIDE,
                BinaryOperator.MODULO
            );
            if (op == null) return lhs;
            lhs = value(binaryOperation(lhs.toValue(), op, parseSelector().toValue()));
        }
    }

    /**
     * <pre>
     * reference-type :=
     *    ( 'boolean' | 'byte' | 'short' | 'int' | 'long' | 'char' | 'float' | 'double' ) '[' ']' { '[' ']' }
     *    | class-or-interface-type { '[' ']' }
     * </pre>
     */
    private Class<?>
    parseReferenceType() throws ParseException {
        Class<?> result;

        if (this.peek(IDENTIFIER) != null) {
            result = parseClassOrInterfaceType();
        } else {
            result = parsePrimitiveType();
            this.read("[");
            this.read("]");
        }

        while (this.peekRead("[")) {
            this.read("]");
            result = Array.newInstance(result, 0).getClass();
        }

        return result;
    }

    private Class<?>
    parsePrimitiveType() throws ParseException {
        switch (read("boolean", "byte", "short", "int", "long", "char", "float", "double")) {
        case 0: return boolean.class;
        case 1: return byte   .class;
        case 2: return short  .class;
        case 3: return int    .class;
        case 4: return long   .class;
        case 5: return char   .class;
        case 6: return float  .class;
        case 7: return double .class;
        default: throw new IllegalStateException();
        }
    }

    /**
     * <pre>
     * selector :=
     *    primary { '.' identifier selector-rest }
     * </pre>
     */
    private Atom<T, E>
    parseSelector() throws ParseException, E {
        Atom<T, E> result = this.parsePrimary();

        if (this.extensions.contains(Extension.NEW_CLASS_WITHOUT_KEYWORD)) {
            List<T> arguments = parseOptionalArguments();
            if (arguments != null) {
                Class<?> clasS = result.toType();
                if (clasS != null) result = value(newClass(clasS, arguments));
            }
        }

        while (peekRead(".")) result = parseSelectorRest(result, read(IDENTIFIER));

        return result;
    }

    /**
     * <pre>
     * selector-rest :=
     *    arguments
     *    | (nothing)
     * </pre>
     */
    private Atom<T, E>
    parseSelectorRest(final Atom<T, E> target, final String identifier) throws ParseException, E {

        {
            final List<T> arguments = parseOptionalArguments();
            if (arguments != null) {

                if (this.extensions.contains(Extension.NEW_CLASS_WITHOUT_KEYWORD)) {
                    String packagE = target.toPackage();
                    if (packagE != null) {
                        final Class<?> clasS = loadClass(packagE + '.' + identifier);
                        if (clasS != null) return new Atom<T, E>() {
                            @Override public T                toValue() throws E { return newClass(clasS, arguments); }
                            @Override public Class<?>         toType()           { return clasS; }
                            @Override public @Nullable String toPackage()        { return null; }
                        };
                    }
                }

                // A method invocation.
                Class<?> clasS = target.toType();
                if (clasS != null) {
                    return value(staticMethodInvocation(clasS, identifier, arguments));
                } else {
                    return value(methodInvocation(target.toValue(), identifier, arguments));
                }
            }
        }

        // A qualified class name?
        {
            String packagE = target.toPackage();
            if (packagE != null) {
                Class<?> clasS = loadClass(packagE + '.' + identifier);
                if (clasS != null) return type(clasS);
            }
        }

        // A static field reference?
        Class<?> type = target.toType();
        if (type != null) return value(staticFieldReference(type, identifier));

        // Must be a package or a nonstatic field reference.
        return new Atom<T, E>() {

            @Override public T
            toValue() throws E, ParseException {
                return fieldReference(target.toValue(), identifier);
            }

            @Override public @Nullable Class<?>
            toType() { return null; }

            @Override public @Nullable String
            toPackage() {
                String packagE = target.toPackage();
                return packagE == null ? null : packagE + '.' + identifier;
            }
        };
    }

    /**
     * <pre>
     * primary-expression :=
     *    '(' expression ')'
     *    | '!' primary-expression
     *    | '-' primary-expression
     *    | string-literal
     *    | int-literal
     *    | long-literal
     *    | float-literal
     *    | double-literal
     *    | 'null' | 'true' | 'false'
     *    | 'new' class-or-interface-type { '[' expression ']' } { '[' ']' }
     *    | 'new' class-or-interface-type arguments
     *    | 'new' primitive-type  '[' expression ']' { '[' expression ']' } { '[' ']' }
     *    | identifier                        // Variable reference
     *    | class-or-interface-type arguments // Class instantiation
     *    | class-or-interface-type           // Class instantiation
     *    | class-or-interface-type           // Class literal
     * </pre>
     */
    private Atom<T, E>
    parsePrimary() throws ParseException, E {

        if (this.peekRead("(")) {
            final Atom<T, E> result = this.parseExpression();
            this.read(")");

            if (this.peek(
                "~", "!", "(",
                IDENTIFIER,
                "this", "new",
                INTEGER_LITERAL, CHARACTER_LITERAL, STRING_LITERAL, FLOATING_POINT_LITERAL
            ) != -1) {
                Class<?> type = result.toType();
                if (type == null) throw new ParseException("'" + type + "' does not pose a type");
                return value(cast(type, this.parsePrimary().toValue()));
            }

            return value(parenthesized(result.toValue()));
        }

        {
            UnaryOperator operator = this.peekReadEnum(
                UnaryOperator.LOGICAL_COMPLEMENT,
                UnaryOperator.MINUS,
                UnaryOperator.BITWISE_COMPLEMENT
            );
            if (operator != null) {
                if (operator == UnaryOperator.MINUS && this.peek(INTEGER_LITERAL) != null) {
                    try {
                        return value(literal(Scanner.decodeIntegerLiteral('-' + this.read().text)));
                    } catch (ScanException se) {
                        throw ExceptionUtil.wrap(null, se, ParseException.class);
                    }
                }
                return value(unaryOperation(operator, this.parseSelector().toValue()));
            }
        }

        if (this.peek(CHARACTER_LITERAL) != null) {
            try {
                return value(literal(Scanner.decodeCharacterLiteral(this.read().text)));
            } catch (ScanException se) {
                throw ExceptionUtil.wrap(null, se, ParseException.class);
            }
        }
        if (this.peek(STRING_LITERAL) != null) {
            try {
                return value(literal(Scanner.decodeStringLiteral(this.read().text)));
            } catch (ScanException se) {
                throw ExceptionUtil.wrap(null, se, ParseException.class);
            }
        }
        if (this.peek(INTEGER_LITERAL) != null) {
            try {
                return value(literal(Scanner.decodeIntegerLiteral(this.read().text)));
            } catch (ScanException se) {
                throw ExceptionUtil.wrap(null, se, ParseException.class);
            }
        }
        if (this.peek(FLOATING_POINT_LITERAL) != null) {
            return value(literal(Scanner.decodeFloatingPointLiteral(this.read().text)));
        }

        if (this.peekRead("null"))  return value(literal(null));
        if (this.peekRead("true"))  return value(literal(Boolean.TRUE));
        if (this.peekRead("false")) return value(literal(Boolean.FALSE));

        if (this.peekRead("new")) {

            // Primitive array creation?
            if (this.peek(IDENTIFIER) == null) {
                return value(this.parseNewArrayRest(this.parsePrimitiveType()));
            }

            Class<?> clasS = this.parseClassOrInterfaceType();

            // Class or interface array creation?
            if (this.peek("[")) return value(parseNewArrayRest(clasS));

            // Class instance creation.
            List<T> arguments = this.parseOptionalArguments();
            if (arguments == null) {
                if (this.extensions.contains(Extension.NEW_CLASS_WITHOUT_PARENTHESES)) {
                    arguments = Collections.emptyList();
                } else {
                    throw new ParseException("Opening parenthesis expected");
                }
            }

            return value(newClass(clasS, arguments));
        }

        final String identifier = this.peekRead(IDENTIFIER);
        if (identifier != null) {

            // An imported type?
            {
                Class<?> clasS = loadImportedClass(identifier);
                if (clasS != null) return type(clasS);
            }

            // Must be a variable reference or a package.
            return new Atom<T, E>() {
                @Override public T                  toValue() throws E, ParseException { return variableReference(identifier); }
                @Override public @Nullable Class<?> toType()                           { return null; }
                @Override public String             toPackage()                        { return identifier; }
            };
        }

        throw new ParseException("Primary expected instead of '" + this.peek() + "'");
    }

    /**
     * <pre>
     *     '[' expression ']' { '[' expression ']' } { '[' ']' }
     * </pre>
     */
    private T
    parseNewArrayRest(Class<?> clasS) throws ParseException, E {
        final List<T> dimensions = new ArrayList<T>();

        read("[");
        dimensions.add(parseExpression().toValue());
        this.read("]");

        while (this.peekRead("[")) {
            if (this.peekRead("]")) {
                clasS = Array.newInstance(clasS, 0).getClass();
                while (this.peekRead("[")) {
                    this.read("]");
                    clasS = Array.newInstance(clasS, 0).getClass();
                }
                break;
            }
            dimensions.add(parseExpression().toValue());
            this.read("]");
        }

        return newArray(clasS, dimensions);
    }

    /**
     * <pre>
     * class-or-interface-type := identifier { '.' identifier }
     * </pre>
     */

    private Class<?>
    parseClassOrInterfaceType() throws ParseException {
        final String identifier = this.read(IDENTIFIER);

        // Imported class.
        Class<?> clasS = this.loadImportedClass(identifier);
        if (clasS != null) return clasS;

        // Qualified class name.
        final List<String> identifiers = new ArrayList<String>();
        identifiers.add(identifier);
        for (;;) {
            clasS = this.loadClass(StringUtil.join(identifiers, "."));
            if (clasS != null) return clasS;

            this.read(".");
            identifiers.add(read(TokenType.IDENTIFIER));
        }
    }

    /**
     * <pre>
     * optional-arguments := [ arguments ]
     *
     * arguments := '(' [ expression { ',' expression } ] ')'
     * </pre>
     *
     * @return {@code null} iff there is no opening parenthesis
     */
    private @Nullable List<T>
    parseOptionalArguments() throws ParseException, E {
        if (!peekRead("(")) return null;
        if (peekRead(")")) return Collections.emptyList();

        List<T> arguments = new ArrayList<T>();
        do {
            arguments.add(this.parseExpression().toValue());
        } while (peekRead(","));
        read(")");
        return arguments;
    }

    // ABSTRACT HANDLERS.

    protected abstract T conditional(T lhs, T mhs, T rhs) throws E;
    protected abstract T unaryOperation(UnaryOperator operator, T operand) throws E;
    protected abstract T binaryOperation(T lhs, BinaryOperator operator, T rhs) throws E;
    protected abstract T fieldReference(T target, String fieldName) throws E;
    protected abstract T staticFieldReference(Class<?> type, String fieldName) throws E;
    protected abstract T methodInvocation(T target, String methodName, List<T> arguments) throws E;
    protected abstract T staticMethodInvocation(Class<?> target, String methodName, List<T> arguments) throws E;
    protected abstract T variableReference(String variableName) throws E, ParseException;
    protected abstract T literal(@Nullable Object value) throws E;
    protected abstract T parenthesized(T value) throws E;
    protected abstract T instanceoF(T lhs, Class<?> rhs) throws E;
    protected abstract T newClass(Class<?> clasS, List<T> arguments) throws E;
    protected abstract T newArray(Class<?> clasS, List<T> dimensions) throws E;
    protected abstract T cast(Class<?> targetType, T operand) throws E, ParseException;

    // HELPERS.

    private @Nullable Class<?>
    loadImportedClass(String simpleClassName) {
        for (String imporT : this.imports) {
            Class<?> clasS = loadClass(imporT + '.' + simpleClassName);
            if (clasS != null) return clasS;
        }
        return null;
    }

    /**
     * @return {@code null} iff a class with the given name cannot be loaded
     */
    private @Nullable Class<?>
    loadClass(String qualifiedClassName) {
        try {
            return this.classLoader.loadClass(qualifiedClassName);
        } catch (ClassNotFoundException cnfe) {
            return null;
        }
    }
}
