
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.text.scanner;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.Charset;

import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.lang.protocol.ProducerWhichThrows;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.scanner.AbstractScanner.Token;

public
class ScannerUtil {

    private ScannerUtil() {}

    /**
     * Converts the {@link StringScanner}, which reads from a {@link CharSequence} into a {@link DocumentScanner},
     * which reads from a {@link Reader} and thinks in "line numbers" and "column numbers").
     *
     * <p>{@link IOException}s thrown by the {@link Reader} are wrapped in {@link ScanException}s.
     */
    public static final <TT extends Enum<TT>> DocumentScanner<TT>
    toDocumentScanner(final StringScanner<TT> stringScanner, Reader reader) {

        @SuppressWarnings("resource") final BufferedReader
        bufferedReader = reader instanceof BufferedReader ? (BufferedReader) reader : new BufferedReader(reader);

        return new DocumentScanner<TT>() {

            private int lineNumber = 0;

            @Override public @Nullable Token<TT>
            produce() throws ScanException {
                for (;;) {
                    Token<TT> token = stringScanner.produce();
                    if (token != null) return token;
                    String line;
                    try {
                        line = bufferedReader.readLine();
                    } catch (IOException ioe) {
                        throw new ScanException(ioe);
                    }
                    if (line == null) return null;
                    this.lineNumber++;
                    stringScanner.setInput(line);
                }
            }

            @Override public int
            getPreviousTokenLineNumber() {
                return this.lineNumber;
            }

            @Override public int
            getPreviousTokenColumnNumber() {
                return stringScanner.getPreviousTokenOffset() + 1;
            }

            @Override public String
            toString() {
                return "Line " + this.getPreviousTokenLineNumber() + ", column " + this.getPreviousTokenColumnNumber();
            }
        };
    }

    /**
     * Uses a given {@link StringScanner} to scan the contents of a given file.
     *
     * @param charset E.g. {@link Charset#forName(String)} or {@link Charset#defaultCharset()}
     * @see           #toDocumentScanner(StringScanner, Reader)
     */
    public static final <TT extends Enum<TT>> ProducerWhichThrows<Token<TT>, ScanException>
    scanner(final StringScanner<TT> stringScanner, File file, Charset charset) throws FileNotFoundException {

        return augmentScanningLocation(toDocumentScanner(
            stringScanner,
            new BufferedReader(new InputStreamReader(new FileInputStream(file), charset))
        ), file.toString());
    }

    public static <T, EX extends Throwable> ProducerWhichThrows<T, EX>
    augmentScanningLocation(final ProducerWhichThrows<T, EX> delegate, final @Nullable String prefix) {
        return new ProducerWhichThrows<T, EX>() {

            @Override public @Nullable T
            produce() throws EX { return delegate.produce(); }

            @Override public @Nullable String
            toString() {
                return prefix == null ? delegate.toString() : prefix + ": " + delegate.toString();
            }
        };
    }

    /**
     * @return A {@link StringScanner} which produces tokens through a {@code delegate}, but only those for which
     *         the {@code predicate} returns {@code true}
     */
    public static <TT extends Enum<TT>> StringScanner<TT>
    filter(final StringScanner<TT> delegate, final Predicate<Token<TT>> predicate) {

        return new StringScanner<TT>() {

            @Override public @Nullable Token<TT>
            produce() throws ScanException {
                for (;;) {
                    Token<TT> token = delegate.produce();
                    if (token == null) return null;
                    if (predicate.evaluate(token)) return token;
                }
            }

            @Override public StringScanner<TT>
            setInput(CharSequence cs) {
                delegate.setInput(cs);
                return this;
            }

            @Override public int
            getOffset() {
                return delegate.getOffset();
            }

            @Override public int
            getPreviousTokenOffset() {
                return delegate.getPreviousTokenOffset();
            }

            @Override public @Nullable String
            toString() { return delegate.toString(); }
        };

    }
}
