
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.util.collections;

import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Set;

import de.unkrig.commons.nullanalysis.NotNullByDefault;

/**
 * An identity map (see {@link java.util.IdentityHashMap}) that performs very well for a SMALL number of entries.
 */
@SuppressWarnings({"unchecked", "rawtypes" })
@NotNullByDefault(false)
public final class SmallIdentityMap<K, V> extends AbstractMap<K, V> {
    Object[] keys;
    Object[] values;
    int size;

    public
    SmallIdentityMap() {
        this.keys = new Object[4];
        this.values = new Object[4];
    }

    @Override public int
    size() {
        return this.size;
    }

    @Override public boolean
    isEmpty() {
        return this.size == 0;
    }

    @Override public boolean
    containsKey(Object key) {
        for (int i = this.size - 1; i >= 0; i--) {
            if (this.keys[i] == key) return true;
        }
        return false;
    }

    @Override public boolean
    containsValue(Object value) {
        for (int i = this.size - 1; i >= 0; i--) {
            if (this.values[i] == value) return true;
        }
        return false;
    }

    @Override public V
    get(Object key) {
        for (int i = this.size - 1; i >= 0; i--) {
            if (this.keys[i] == key) return (V) this.values[i];
        }
        return null;
    }

    @Override public V
    put(K key, V value) {
        for (int i = this.size - 1; i >= 0; i--) {
            if (this.keys[i] == key) {
                Object oldValue = this.values[i];
                this.values[i] = value;
                return (V) oldValue;
            }
        }
        int n = this.size;
        if (n == this.keys.length) {
            Object[] tmpKeys = new Object[n << 1];
            System.arraycopy(this.keys, 0, tmpKeys, 0, n);
            this.keys = tmpKeys;
            Object[] tmpValues = new Object[n << 1];
            System.arraycopy(this.values, 0, tmpValues, 0, n);
            this.values = tmpValues;
        }
        this.keys[n] = key;
        this.values[n] = value;
        this.size = n + 1;
        return null;
    }

    @Override public V
    remove(Object key) {
        for (int i = this.size - 1; i >= 0; i--) {
            if (this.keys[i] == key) {
                Object result = this.values[i];
                int n = this.size - 1;
                for (; i < n; i++) {
                    this.keys[i] = this.keys[i + 1];
                    this.values[i] = this.values[i + 1];
                }
                this.keys[n] = null;
                this.values[n] = null;
                this.size = n;
                return (V) result;
            }
        }
        return null;
    }

    @Override public void
    clear() {
        for (int i = this.size - 1; i >= 0; i--) {
            this.keys[i] = null;
            this.values[i] = null;
        }
        this.size = 0;
    }

    @Override public Set<Entry<K, V>>
    entrySet() {
        if (this.entrySet == null) {
            this.entrySet = new EntrySet();
        }
        return this.entrySet();
    }
    private Object entrySet;

    @NotNullByDefault(false) final class EntrySet extends AbstractSet<Entry<K, V>> {

        @Override public int
        size() {
            return SmallIdentityMap.this.size;
        }

        @Override public boolean
        isEmpty() {
            return SmallIdentityMap.this.size == 0;
        }

        @Override public boolean
        contains(Object object) {
            if (!(object instanceof Entry)) return false;
            Entry entry = (Entry) object;
            for (int i = SmallIdentityMap.this.size - 1; i >= 0; i--) {
                if (SmallIdentityMap.this.keys[i] == entry.getKey()) {
                    Object v = SmallIdentityMap.this.values[i];
                    if (v == null ? entry.getValue() == null : v.equals(entry.getValue())) {
                        return true;
                    }
                }
            }
            return false;
        }

        @Override public Iterator<Entry<K, V>>
        iterator() {
            return new Iterator<Entry<K,V>>() {

                private int pos = SmallIdentityMap.this.size;

                @Override public boolean
                hasNext() {
                    return this.pos > 0;
                }

                @Override public Entry<K, V>
                next() {
                    if (this.pos == 0) throw new NoSuchElementException();
                    final int n = --this.pos;
                    final Object key = SmallIdentityMap.this.keys[n];
                    final Object[] values = SmallIdentityMap.this.values;
                    return newEntry(key, values, n);
                }

                @Override public void
                remove() {
                    if (this.pos == SmallIdentityMap.this.size) throw new IllegalStateException();
                    int n = SmallIdentityMap.this.size - 1;
                    for (int i = this.pos; i < n; i++) {
                        SmallIdentityMap.this.keys[i] = SmallIdentityMap.this.keys[i + 1];
                        SmallIdentityMap.this.values[i] = SmallIdentityMap.this.values[i + 1];
                    }
                    SmallIdentityMap.this.keys[n] = null;
                    SmallIdentityMap.this.values[n] = null;
                    SmallIdentityMap.this.size = n;
                }
            };
        }

        @Override public Object[]
        toArray() {
            int n = SmallIdentityMap.this.size;
            Object[] result = new Object[n];
            Object[] keys = SmallIdentityMap.this.keys;
            Object[] values = SmallIdentityMap.this.values;
            for (int i = n - 1; i >= 0; i--) {
                result[i] = newEntry(keys[i], values, i);
            }
            return result;
        }

        @Override public <T> T[]
        toArray(T[] result) {
            int n = SmallIdentityMap.this.size;
            if (result.length < n) {
                result = (T[]) new Entry[n];
            }
            Object[] keys = SmallIdentityMap.this.keys;
            Object[] values = SmallIdentityMap.this.values;
            for (int i = n - 1; i >= 0; i--) {
                result[i] = (T) newEntry(keys[i], values, i);
            }
            return result;
        }

        @Override public boolean
        add(Entry<K, V> entry) {
            return SmallIdentityMap.this.put(entry.getKey(), entry.getValue()) != null;
        }

        @Override public boolean
        remove(Object key) {
            return SmallIdentityMap.this.remove(key) != null;
        }

        @Override public void
        clear() {
            SmallIdentityMap.this.clear();
        }

        Entry<K, V>
        newEntry(final Object key, final Object[] values, final int idx) {
            final Object value = values[idx];
            return new Entry<K, V>() {
                @Override public K getKey()   { return (K) key; }
                @Override public V getValue() { return (V) value; }

                @Override public V
                setValue(Object newValue) {
                    Object result = values[idx];
                    values[idx] = newValue;
                    return (V) result;
                }
            };
        }
    }
}
